# Copyright 2022 G-Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import threading

from horovod.runner.common.util import timeout, network
from horovod.runner.common.util.network import AckResponse
from horovod.runner.common.util.timeout import TimeoutException
from horovod.runner.util.threads import in_thread

"""
Items sent over the Wire between Compute Client and Service.
"""


class RegisterDispatcherRequest(object):
    """Registers a dispatcher server address along with a dispatcher id."""

    def __init__(self, dispatcher_id, dispatcher_address):
        self.dispatcher_id = dispatcher_id
        """Id of the dispatcher server (0 indexed)"""

        self.dispatcher_address = dispatcher_address
        """Address of the dispatcher"""


class WaitForDispatcherRegistrationRequest(object):
    """Wait for the given dispatcher to register. Blocks until the dispatcher registers or the timeout occurs."""

    def __init__(self, dispatcher_id, timeout):
        self.dispatcher_id = dispatcher_id
        """Dispatcher id"""

        self.timeout = timeout
        """Wait timeout in seconds"""


class WaitForDispatcherRegistrationResponse(object):
    """Response that the dispatcher has registered, providing its address."""
    def __init__(self, dispatcher_address):
        self.dispatcher_address = dispatcher_address
        """Address of the requested dispatcher."""


class RegisterDispatcherWorkerRequest(object):
    """Registers a worker server for a dispatcher server."""

    def __init__(self, dispatcher_id, worker_id):
        self.dispatcher_id = dispatcher_id
        """Id of the dispatcher server (0 indexed)"""

        self.worker_id = worker_id
        """Id of the worker server (0 indexed)"""


class WaitForDispatcherWorkerRegistrationRequest(object):
    """Wait for all workers of the given dispatcher to register.
    Blocks until all workers registers or the timeout occurs."""

    def __init__(self, dispatcher_id, timeout):
        self.dispatcher_id = dispatcher_id
        """Dispatcher id"""

        self.timeout = timeout
        """Wait timeout in seconds"""


class ShutdownRequest(object):
    """Initiate the shutdown of the compute service as it is no longer needed."""


class WaitForShutdownRequest(object):
    """Wait for the compute service to shutdown. Blocks until the shutdown is initiated."""


"""
ComputeService is used to communicate between training driver, training tasks, compute driver and compute tasks.
It is ML framework agnostic, though currently only used for Tensorflow data service (tf.data.experimental.service).
It orchestrates synchronization between tf.DispatchServer, tf.WorkerServer, training and compute tasks.

ComputeClient is used to query and change the internal state of the ComputeService.
"""


class ComputeService(network.BasicService):
    NAME = "Compute service"

    def __init__(self, dispatchers, workers_per_dispatcher, key, nics=None):
        if dispatchers <= 0:
            raise ValueError(f'The number of dispatchers must be larger than 0: {dispatchers}')
        if workers_per_dispatcher <= 0:
            raise ValueError(f'The number of workers per dispatcher must be larger than 0: {workers_per_dispatcher}')

        self._max_dispatcher_id = dispatchers - 1
        self._dispatcher_addresses = [None] * dispatchers
        self._workers_per_dispatcher = workers_per_dispatcher
        self._dispatcher_worker_ids = [set()] * dispatchers
        self._shutdown = False
        self._wait_cond = threading.Condition()

        super().__init__(ComputeService.NAME, key, nics)

    def _handle(self, req, client_address):
        if isinstance(req, RegisterDispatcherRequest):
            self._wait_cond.acquire()
            try:
                if not 0 <= req.dispatcher_id <= self._max_dispatcher_id:
                    return IndexError(f'Dispatcher id must be within [0..{self._max_dispatcher_id}]: '
                                      f'{req.dispatcher_id}')

                if self._dispatcher_addresses[req.dispatcher_id] is not None and \
                   self._dispatcher_addresses[req.dispatcher_id] != req.dispatcher_address:
                    return ValueError(f'Dispatcher with id {req.dispatcher_id} has already been registered under '
                                      f'different address {self._dispatcher_addresses[req.dispatcher_id]}: '
                                      f'{req.dispatcher_address}')

                self._dispatcher_addresses[req.dispatcher_id] = req.dispatcher_address
                self._wait_cond.notify_all()
            finally:
                self._wait_cond.release()
            return network.AckResponse()

        if isinstance(req, WaitForDispatcherRegistrationRequest):
            self._wait_cond.acquire()
            try:
                if not 0 <= req.dispatcher_id <= self._max_dispatcher_id:
                    return IndexError(f'Dispatcher id must be within [0..{self._max_dispatcher_id}]: '
                                      f'{req.dispatcher_id}')

                tmout = timeout.Timeout(timeout=req.timeout,
                                        message='Timed out waiting for {activity}. Try to find out what takes '
                                                'the dispatcher so long to register or increase timeout.')

                while self._dispatcher_addresses[req.dispatcher_id] is None:
                    self._wait_cond.wait(tmout.remaining())
                    tmout.check_time_out_for(f'dispatcher {req.dispatcher_id} to register')
            except TimeoutException as e:
                return e
            finally:
                self._wait_cond.release()
            return WaitForDispatcherRegistrationResponse(self._dispatcher_addresses[req.dispatcher_id])

        if isinstance(req, RegisterDispatcherWorkerRequest):
            self._wait_cond.acquire()
            try:
                if not 0 <= req.dispatcher_id <= self._max_dispatcher_id:
                    return IndexError(f'Dispatcher id must be within [0..{self._max_dispatcher_id}]: '
                                      f'{req.dispatcher_id}')

                self._dispatcher_worker_ids[req.dispatcher_id].update({req.worker_id})
                self._wait_cond.notify_all()
            finally:
                self._wait_cond.release()
            return network.AckResponse()

        if isinstance(req, WaitForDispatcherWorkerRegistrationRequest):
            # if there is only a single dispatcher, wait for that one instead of the requested one
            dispatcher_id = req.dispatcher_id if self._max_dispatcher_id > 0 else 0

            self._wait_cond.acquire()
            try:
                if not 0 <= req.dispatcher_id <= self._max_dispatcher_id:
                    return IndexError(f'Dispatcher id must be within [0..{self._max_dispatcher_id}]: '
                                      f'{req.dispatcher_id}')

                tmout = timeout.Timeout(timeout=req.timeout,
                                        message='Timed out waiting for {activity}. Try to find out what takes '
                                                'the workers so long to register or increase timeout.')

                while len(self._dispatcher_worker_ids[dispatcher_id]) < self._workers_per_dispatcher:
                    self._wait_cond.wait(tmout.remaining())
                    tmout.check_time_out_for(f'workers for dispatcher {dispatcher_id} to register')
            except TimeoutException as e:
                return e
            finally:
                self._wait_cond.release()
            return network.AckResponse()

        if isinstance(req, ShutdownRequest):
            in_thread(self.shutdown)
            return network.AckResponse()

        if isinstance(req, WaitForShutdownRequest):
            self._wait_cond.acquire()
            try:
                while not self._shutdown:
                    self._wait_cond.wait()
            finally:
                self._wait_cond.release()
            return network.AckResponse()

        return super()._handle(req, client_address)

    def shutdown(self):
        self._wait_cond.acquire()
        try:
            # notify all requests that are waiting for shutdown
            self._shutdown = True
            self._wait_cond.notify_all()
        finally:
            self._wait_cond.release()

        # this will wait until all requests have been completed
        super(ComputeService, self).shutdown()


class ComputeClient(network.BasicClient):

    def __init__(self, compute_addresses, key, verbose=1):
        super().__init__(ComputeService.NAME, compute_addresses, key, verbose)

    def register_dispatcher(self, dispatcher_id, dispatcher_address):
        self._send(RegisterDispatcherRequest(dispatcher_id, dispatcher_address))

    def wait_for_dispatcher_registration(self, dispatcher_id, timeout) -> str:
        resp = self._send(WaitForDispatcherRegistrationRequest(dispatcher_id, timeout))
        return resp.dispatcher_address

    def register_worker_for_dispatcher(self, dispatcher_id, worker_id):
        self._send(RegisterDispatcherWorkerRequest(dispatcher_id, worker_id))

    def wait_for_dispatcher_worker_registration(self, dispatcher_id, timeout):
        self._send(WaitForDispatcherWorkerRegistrationRequest(dispatcher_id, timeout))

    def shutdown(self):
        self._send(ShutdownRequest())

    def wait_for_shutdown(self):
        self._send(WaitForShutdownRequest())

    def _send(self, req, stream=None):
        """Raise exceptions that we retrieve for any request."""
        resp = super(ComputeClient, self)._send(req, stream)
        if isinstance(resp, Exception):
            raise resp
        return resp



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
# Modifications copyright Microsoft
# Modifications copyright (C) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import atexit
import io
import os
import re
import shutil
import subprocess
import sys
import tempfile
import textwrap

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext

from horovod import __version__

_FRAMEWORK_METADATA_FILE = 'horovod/metadata.json'

class CMakeExtension(Extension):
    def __init__(self, name, cmake_lists_dir='.', sources=None, **kwa):
        if sources is None:
            sources = []
        Extension.__init__(self, name, sources=sources, **kwa)
        self.cmake_lists_dir = os.path.abspath(cmake_lists_dir)


tensorflow_mpi_lib = CMakeExtension('horovod.tensorflow.mpi_lib',
                                    cmake_lists_dir='.', sources=[])
torch_mpi_lib_v2 = CMakeExtension('horovod.torch.mpi_lib_v2',
                                  cmake_lists_dir='.', sources=[])
mxnet_mpi_lib = CMakeExtension('horovod.mxnet.mpi_lib',
                               cmake_lists_dir='.', sources=[])

def is_build_action():
    if len(sys.argv) <= 1:
        return False

    if sys.argv[1].startswith('build'):
        return True

    if sys.argv[1].startswith('bdist'):
        return True

    if sys.argv[1].startswith('install'):
        return True

    if sys.argv[1].startswith('develop'):
        return True

def get_cmake_bin():
    from packaging import version

    if 'HOROVOD_CMAKE' in os.environ:
        return os.environ['HOROVOD_CMAKE']

    cmake_bin = 'cmake'
    try:
        out = subprocess.check_output([cmake_bin, '--version'])
    except OSError:
        cmake_installed_version = version.parse("0.0")
    else:
        cmake_installed_version = version.parse(re.search(r'version\s*([\d.]+)', out.decode()).group(1))

    if cmake_installed_version < version.parse("3.13.0"):
        print("Could not find a recent CMake to build Horovod. "
              "Attempting to install CMake 3.13 to a temporary location via pip.", flush=True)
        cmake_temp_dir = tempfile.TemporaryDirectory(prefix="horovod-cmake-tmp")
        atexit.register(cmake_temp_dir.cleanup)
        try:
            _ = subprocess.check_output(["pip", "install", "--target", cmake_temp_dir.name, "cmake~=3.13.0"])
        except Exception:
            raise RuntimeError("Failed to install temporary CMake. "
                               "Please update your CMake to 3.13+ or set HOROVOD_CMAKE appropriately.")
        cmake_bin = os.path.join(cmake_temp_dir.name, "bin", "run_cmake")
        with io.open(cmake_bin, "w") as f_run_cmake:
            f_run_cmake.write(
                f"#!/bin/sh\nPYTHONPATH={cmake_temp_dir.name} {os.path.join(cmake_temp_dir.name, 'bin', 'cmake')} \"$@\"")
        os.chmod(cmake_bin, 0o755)

    return cmake_bin


class custom_build_ext(build_ext):
    def build_extensions(self):
        if os.getenv('HOROVOD_SKIP_COMPILE') == '1':
            # Skip building extensions using CMake
            print("Horovod is being installed without native libraries")
            return

        cmake_bin = get_cmake_bin()

        config = 'Debug' if self.debug or os.environ.get('HOROVOD_DEBUG') == "1" else 'RelWithDebInfo'

        ext_name = self.extensions[0].name
        build_dir = self.get_ext_fullpath(ext_name).replace(self.get_ext_filename(ext_name), '')
        build_dir = os.path.abspath(build_dir)

        cmake_args = ['-DCMAKE_BUILD_TYPE=' + config,
                      '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(config.upper(), build_dir),
                      '-DPYTHON_EXECUTABLE:FILEPATH=' + sys.executable]

        make_args = ['-j8'] if not os.environ.get('MAKEFLAGS') else []
        if self.verbose:
            make_args.append('VERBOSE=1')

        cmake_build_args = ['--config', config]
        if make_args:
            # -- specifies that these args are going to the native build tool: make
            cmake_build_args += ['--'] + make_args

        cmake_build_dir = os.path.join(self.build_temp, config)
        if not os.path.exists(cmake_build_dir):
            os.makedirs(cmake_build_dir)

        config_and_build_commands = [
            [cmake_bin, self.extensions[0].cmake_lists_dir] + cmake_args,
            [cmake_bin, '--build', '.'] + cmake_build_args
        ]

        if self.verbose:
            print(f"Running CMake in {cmake_build_dir}:")
            for command in config_and_build_commands:
                print(" ".join(command))
            sys.stdout.flush()

        # Config and build the extension
        try:
            for command in config_and_build_commands:
                subprocess.check_call(command, cwd=cmake_build_dir)
        except OSError as e:
            raise RuntimeError('CMake failed: {}'.format(str(e)))

        if sys.argv[1].startswith('develop'):
            # Copy over metadata.json file from build directory
            shutil.copyfile(os.path.join(build_dir, _FRAMEWORK_METADATA_FILE),
                            os.path.join(self.extensions[0].cmake_lists_dir, _FRAMEWORK_METADATA_FILE))
            # Remove unfound frameworks, otherwise develop mode will fail the install
            self.extensions = [x for x in self.extensions if os.path.exists(self.get_ext_fullpath(x.name))]


# python packages required to use horovod in general
require_list = ['cloudpickle', 'psutil', 'pyyaml', 'dataclasses;python_version<"3.7"', 'packaging']

# framework dependencies
tensorflow_require_list = ['tensorflow']
tensorflow_cpu_require_list = ['tensorflow-cpu']
tensorflow_gpu_require_list = ['tensorflow-gpu']
keras_require_list = ['keras>=2.0.8,!=2.0.9,!=2.1.0,!=2.1.1']
# pytorch-lightning 1.3.8 is a stable version to work with horovod
pytorch_require_list = ['torch']
mxnet_require_list = ['mxnet>=1.4.1']
pyspark_require_list = ['pyspark>=2.3.2;python_version<"3.8"',
                        'pyspark>=3.0.0;python_version>="3.8"']
spark_require_list = ['numpy', 'petastorm>=0.12.0', 'pyarrow>=0.15.0,<11.0', 'fsspec>=2021.07.0']
# https://github.com/ray-project/ray/pull/17465
# google-api-core>=2.9.0 depends on protobuf<5.0.0dev,>=3.20.1, which conflicts with
#   tensorflow protobuf~=3.20 and pytorch-lightning protobuf<3.20,>=3.9.2
ray_require_list = ['ray', 'aioredis<2', 'google-api-core<2.9.0']
pytorch_spark_require_list = pytorch_require_list + \
                             spark_require_list + \
                             pyspark_require_list + \
                             ['pytorch_lightning>=1.3.8,<1.5.10']

# all frameworks' dependencies
all_frameworks_require_list = tensorflow_require_list + \
                              keras_require_list + \
                              pytorch_require_list + \
                              mxnet_require_list + \
                              spark_require_list + \
                              pyspark_require_list

# python packages required / recommended to develop horovod
# these are the earliest versions to work with Python 3.8
# keep in sync with Dockerfile.test.cpu
# NOTE: do not use versions with +cpu or +gpu here as users would need to add --find-links to pip
dev_require_list = ['tensorflow-cpu==2.2.0',
                    'keras==2.3.1',
                    'torch==1.4.0',
                    'torchvision==0.5.0',
                    'pytorch_lightning>=1.3.8,<1.5.10',
                    'mxnet==1.5.0',
                    'pyspark==3.0.1'] + spark_require_list
# torchvision 0.5.0 depends on torch==1.4.0

# python packages required only to run tests
test_require_list = ['mock', 'pytest', 'pytest-forked', 'pytest-subtests', 'parameterized']

# Skip cffi if pytorch extension explicitly disabled
if not os.environ.get('HOROVOD_WITHOUT_PYTORCH'):
    require_list.append('cffi>=1.4.0')


def get_package_version():
    return __version__ + "+" + os.environ['HOROVOD_LOCAL_VERSION'] if 'HOROVOD_LOCAL_VERSION' in os.environ else __version__


setup(name='horovod',
      version=get_package_version(),
      packages=find_packages(),
      description='Distributed training framework for TensorFlow, Keras, PyTorch, and Apache MXNet.',
      author='The Horovod Authors',
      license='Apache 2.0',
      long_description=textwrap.dedent('''\
          Horovod is a distributed training framework for TensorFlow, Keras, PyTorch, and Apache MXNet.
          The goal of Horovod is to make distributed Deep Learning fast and easy to use.'''),
      url='https://github.com/horovod/horovod',
      keywords=['deep learning', 'tensorflow', 'keras', 'pytorch', 'mxnet', 'spark', 'AI'],
      classifiers=[
          'License :: OSI Approved :: Apache Software License',
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      ext_modules=[tensorflow_mpi_lib, torch_mpi_lib_v2, mxnet_mpi_lib],
      cmdclass={'build_ext': custom_build_ext},
      # cffi is required for PyTorch
      # If cffi is specified in setup_requires, it will need libffi to be installed on the machine,
      # which is undesirable.  Luckily, `install` action will install cffi before executing build,
      # so it's only necessary for `build*` or `bdist*` actions.
      setup_requires=require_list if is_build_action() else [],
      install_requires=require_list,
      tests_require=test_require_list,
      extras_require={
          'all-frameworks': all_frameworks_require_list,
          'tensorflow': tensorflow_require_list,
          'tensorflow-cpu': tensorflow_cpu_require_list,
          'tensorflow-gpu': tensorflow_gpu_require_list,
          'keras': keras_require_list,
          'pytorch': pytorch_require_list,
          'mxnet': mxnet_require_list,
          'spark': spark_require_list + pyspark_require_list,
          'pytorch-spark': pytorch_spark_require_list,
          'ray': ray_require_list,
          'dev': dev_require_list,
          'test': test_require_list,
      },
      python_requires='>=3.6',
      zip_safe=False,
      entry_points={
          'console_scripts': [
              'horovodrun = horovod.runner.launch:run_commandline'
          ]
      })



import os
import tensorflow as tf
from packaging import version
from horovod.tensorflow.mpi_ops import size_op
from horovod.tensorflow.mpi_ops import global_process_set


_POST_TF_2_4_0 = version.parse(tf.__version__) >= version.parse('2.4.0')
_IS_TF2 = version.parse(tf.__version__) >= version.parse('2.0.0')


class LocalGradientAggregationHelperEager:
    """
   LocalGradientAggregationHelperEager aggregates gradient updates
   locally, and communicates the updates across machines only once
   every backward_passes_per_step. Only supports eager execution.
   """

    def __init__(
        self,
        backward_passes_per_step,
        allreduce_func,
        sparse_as_dense,
        average_aggregated_gradients,
        process_set=global_process_set,
        scale_local_gradients=True
    ):
        self.allreduce_grads = allreduce_func
        self.sparse_as_dense = sparse_as_dense

        # backward_passes_per_step controls how often gradient updates are
        # synchronized.
        self.backward_passes_per_step = backward_passes_per_step
        if self.backward_passes_per_step <= 0:
            raise ValueError("backward_passes_per_step must be > 0")

        # average_aggregated_gradients controls whether gradient updates that are
        # aggregated, should be divided by `backward_passes_per_step`.
        self.average_aggregated_gradients = average_aggregated_gradients

        # This is going to be N data structure holding the aggregated gradient updates
        # for parameter updates. N is the number of parameters.
        self.locally_aggregated_grads = {}

        # Used to know when to allreduce and apply gradients. We allreduce when `self.counter`
        # is equal to `self.backward_passes_per_step`. We apply gradients when `self.counter`
        # is equal to 0.
        self.counter = tf.Variable(initial_value=0)

        self.process_set = process_set
        self.scale_local_gradients = scale_local_gradients
        self._local_vars = set()

    def register_local_var(self, var):
        """Registers a source/variable as worker local. Horovod will not perform any global
        operations on gradients corresponding to these sources and will instead return the local
        gradient."""
        if _IS_TF2:
            self._local_vars.add(var.ref())
        else:
            self._local_vars.add(var)

    def compute_gradients(self, grads, vars):
        # On steps where allreduce happens, resulting_grads returns the allreduced
        # gradients, on other steps it returns the locally aggregated
        # gradients.
        resulting_grads = []

        for idx, grad in enumerate(grads):
            # Handle IndexedSlices.
            if self.sparse_as_dense and isinstance(grad, tf.IndexedSlices):
                grad = tf.convert_to_tensor(grad)
            elif isinstance(grad, tf.IndexedSlices):
                raise ValueError(
                    "IndexedSlices are not supported when "
                    "`backward_passes_per_step` > 1 and "
                    "`sparse_as_dense` is False."
                )

            # Create variables to store to aggregate gradients if they don't
            # already exist. Skip variables that are None.
            if idx not in self.locally_aggregated_grads.keys():
                if grad is not None:
                    self.locally_aggregated_grads[idx] = tf.Variable(
                        initial_value=tf.zeros_like(grad),
                        trainable=False,
                        dtype=grad.dtype,
                    )

            if grad is None:
                resulting_grads.append(None)
            else:
                self.locally_aggregated_grads[idx].assign_add(grad)
                resulting_grads.append(
                    self.locally_aggregated_grads[idx].read_value())
        assert len(self.locally_aggregated_grads) == len(grads)

        # Increment counter.
        self.counter.assign_add(1)

        def _all_reduce_and_clear_aggregated_variables(aggregated_gradients, vars):
            # Performs allreduce. If `average_aggregated_gradients` is
            # set to True divides result by `backward_passes_per_step`.
            reduced_gradients = self._allreduce_helper(aggregated_gradients, vars)
            assert len(reduced_gradients) == len(grads)

            self._clear_vars()
            return reduced_gradients

        def _do_nothing(aggregated_gradients):
            return aggregated_gradients

        resulting_grads = tf.cond(
            pred=tf.equal(self.counter, self.backward_passes_per_step),
            true_fn=lambda: _all_reduce_and_clear_aggregated_variables(resulting_grads, vars),
            false_fn=lambda: _do_nothing(resulting_grads),
        )

        return resulting_grads

    def _allreduce_helper(self, grads, vars):
        def __filtered_reduce_grads(grads, vars):
            rv = []
            rg = []
            if _IS_TF2:
                v2g = {var.ref(): grad for var, grad in zip(vars, grads)}
                for var, grad in zip(vars, grads):
                    if var.ref() not in self._local_vars:
                        rv.append(var)
                        rg.append(grad)
            else:
                v2g = {var: grad for var, grad in zip(vars, grads)}
                for var, grad in zip(vars, grads):
                    if var not in self._local_vars:
                        rv.append(var)
                        rg.append(grad)

            rg = self.allreduce_grads(rg, rv)
            horovod_size = size_op(process_set_id=self.process_set.process_set_id) if int(os.environ.get("HOROVOD_ELASTIC", 0)) else self.process_set.size()
            if _IS_TF2:
                for rv, rg in zip(rv, rg):
                    v2g[rv.ref()] = rg

                if self.scale_local_gradients and len(self._local_vars):
                    # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
                    for v_ref in v2g:
                        if v_ref in self._local_vars and v2g[v_ref] is not None:
                            v2g[v_ref] /= horovod_size

                return [v2g[rv.ref()] for rv in vars]
            else:
                for rv, rg in zip(rv, rg):
                    v2g[rv] = rg

                if self.scale_local_gradients and len(self._local_vars):
                    # Scale local gradients by a size factor. See pull/3695 and discussions/3705 for context.
                    for v in v2g:
                        if v in self._local_vars and v2g[v] is not None:
                            v2g[v] /= horovod_size

                return [v2g[rv] for rv in vars]

        allreduced_grads = __filtered_reduce_grads(grads, vars)

        if not self.average_aggregated_gradients:
            return allreduced_grads

        scaled_allreduced_grads = []
        for grad in allreduced_grads:
            if grad is None:
                scaled_allreduced_grads.append(grad)
                continue

            scaled_allreduced_grads.append(
                grad / self.backward_passes_per_step)

        return scaled_allreduced_grads

    def _clear_vars(self):
        self.counter.assign(0)
        for idx in self.locally_aggregated_grads.keys():
            self.locally_aggregated_grads[idx].assign(
                tf.zeros_like(self.locally_aggregated_grads[idx]))

    def apply_gradients(self, apply_grads_closure, optimizer, *args, **kwargs):
        def increment_optimizer_iteration():
            # (kvignesh1420): Since all `tf.OptimizerV2` instances have the `iterations`
            # property for modifying the underlying `optimizer._iterations`, it is safe to use
            # the property instead of the private variable. For instance, the keras
            # `LossScaleOptimizer` inherits `tf.Optimizer` and exposes the cleaner `iterations`
            # property instead of the unsafe `_iterations`.

            if hasattr(optimizer, "iterations") and optimizer.iterations is not None:
                if hasattr(super(optimizer.__class__, optimizer), '_compute_gradients'):
                    return optimizer.iterations.assign_add(1).op
                else:
                    return optimizer.iterations.assign_add(1)
            return tf.no_op()

        def non_aggregation_step():
            if _POST_TF_2_4_0:
                # In TF 2.4+ `_aggregate_gradients()` is called from inside of `apply_gradients()`.
                # We account for this by calling `_aggregate_gradients()` for steps where we do
                # not call `apply_gradients()`.
                transformed_grads_and_vars = optimizer._transform_unaggregated_gradients(
                    args[0]) if hasattr(optimizer, '_transform_unaggregated_gradients') else args[0]
                _ = optimizer._aggregate_gradients(transformed_grads_and_vars)

            return increment_optimizer_iteration()

        def is_aggregation_step():
            return tf.equal(self.counter, 0)

        return tf.cond(
            pred=is_aggregation_step(),
            true_fn=apply_grads_closure,
            false_fn=non_aggregation_step,
        )



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from petastorm.pytorch import BatchedDataLoader, InMemBatchedDataLoader
from horovod.data import BaseDataLoader, AsyncDataLoaderMixin


class PytorchDataLoader(BaseDataLoader):
    def __init__(self, reader, batch_size, shuffling_queue_capacity, name="",
                 limit_step_per_epoch=-1, verbose=False):
        self.reader = reader
        self.batch_size = batch_size
        self.shuffling_queue_capacity = shuffling_queue_capacity
        self.limit_step_per_epoch = limit_step_per_epoch
        self.name = name
        self.verbose = verbose

        print(f"[{self.name}]: Initializing petastorm dataloader with batch_size={batch_size}"
              f"shuffling_queue_capacity={shuffling_queue_capacity}, "
              f"limit_step_per_epoch={limit_step_per_epoch}")

    def __len__(self):
        # We cannot infer length from reader.
        return self.limit_step_per_epoch if self.limit_step_per_epoch != -1 else 0

    def _iterate(self):
        # Reset the reader if needed.
        if self.reader.last_row_consumed:
            self._print_verbose(f"[{self.name}]: Resetting Petastorm reader for {self.reader.dataset.paths}")
            self.reader.reset()

        # Re-create the data loader for each iteration. This is needed becasue there may be
        # some left-over data from last epoch which can cause petastorm's BatchedDataLoader
        # fail to start new iteration. To workaround the issue, we have to re-create the data
        # loader at each new iterration starts.
        data_loader = BatchedDataLoader(
            self.reader,
            batch_size=self.batch_size,
            shuffling_queue_capacity=self.shuffling_queue_capacity,
        )

        num_steps = 0

        self._print_verbose(f"[{self.name}]: Start to generate batch data. limit_step_per_epoch={self.limit_step_per_epoch}")

        for batch in data_loader:
            if num_steps == self.limit_step_per_epoch:
                self._print_verbose(f"[{self.name}]: Reach limit_step_per_epoch. Stop at step {num_steps}.")
                break

            num_steps += 1
            yield batch

    def _print_verbose(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)


class PytorchAsyncDataLoader(AsyncDataLoaderMixin, PytorchDataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class PytorchInfiniteDataLoader(PytorchDataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.reader.num_epochs is not None:
            raise ValueError("Need to set num_epochs as None in reader.")

        self.data_loader = BatchedDataLoader(
            self.reader,
            batch_size=self.batch_size,
            shuffling_queue_capacity=self.shuffling_queue_capacity)
        self.iterator = iter(self.data_loader)

    def _iterate(self):
        num_steps = 0
        self._print_verbose(f"[{self.name}]: Start to generate batch data. limit_step_per_epoch={self.limit_step_per_epoch}")

        while True:
            if num_steps == self.limit_step_per_epoch:
                self._print_verbose(f"[{self.name}]: Reach limit_step_per_epoch. Stop at step {num_steps}.")
                break
            num_steps += 1

            yield next(self.iterator)


class PytorchInfiniteAsyncDataLoader(AsyncDataLoaderMixin, PytorchInfiniteDataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class PytorchInmemDataLoader(BaseDataLoader):
    def __init__(self, reader, batch_size, num_epochs, name="",
                 shuffle=False, limit_step_per_epoch=-1, verbose=False):
        self.batch_size = batch_size
        self.limit_step_per_epoch = limit_step_per_epoch
        self.name = name
        self.verbose = verbose

        if limit_step_per_epoch == -1:
            raise ValueError('limit_step_per_epoch cannot be -1 for inmem dataloader')

        print(f"[{self.name}]: Initializing petastorm inmem_dataloader with batch_size={batch_size}"
              f"num_epochs={num_epochs}, "
              f"shuffle={shuffle}"
              f"limit_step_per_epoch={limit_step_per_epoch}")

        self.dataloader = InMemBatchedDataLoader(reader, batch_size=batch_size, num_epochs=num_epochs,
                                                 rows_capacity=batch_size*limit_step_per_epoch, shuffle=shuffle)
        self.iterator = iter(self.dataloader)

    def __len__(self):
        # We cannot infer length from reader.
        return self.limit_step_per_epoch

    def _iterate(self):
        num_steps = 0
        self._print_verbose(f"[{self.name}]: Start to generate batch data. limit_step_per_epoch={self.limit_step_per_epoch}")

        while True:
            if num_steps == self.limit_step_per_epoch:
                self._print_verbose(f"[{self.name}]: Reach limit_step_per_epoch. Stop at step {num_steps}.")
                break
            num_steps += 1

            yield next(self.iterator)

    def _print_verbose(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)


class PytorchInmemAsyncDataLoader(AsyncDataLoaderMixin, PytorchInmemDataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class PetastormBatchedDataLoader(BatchedDataLoader):
    def __init__(self, name="", limit_step_per_epoch=-1, verbose=False, *args, **kwargs):
        print(f"[{name}]Petastorm BatchedDataLoader will ignore limit_step_per_epoch and verbose.")
        super().__init__(*args, **kwargs)



from collections import defaultdict
import logging
from typing import Dict

import ray
from ray.util.placement_group import get_current_placement_group
from horovod.ray.utils import map_blocking
from horovod.ray.worker import BaseHorovodWorker

logger = logging.getLogger(__name__)


def create_placement_group(resources_per_bundle: Dict[str, int],
                           num_bundles: int, pg_timeout: int,
                           pg_strategy: str):
    bundles = [resources_per_bundle.copy() for _ in range(num_bundles)]
    pg = ray.util.placement_group(bundles, strategy=pg_strategy)
    logger.debug("Waiting for placement group to start.")
    ready, _ = ray.wait([pg.ready()], timeout=pg_timeout)
    if ready:
        logger.debug("Placement group has started.")
    else:
        raise TimeoutError("Placement group creation timed out. Make sure "
                           "your cluster either has enough resources or use "
                           "an autoscaling cluster. Current resources "
                           "available: {}, resources requested by the "
                           "placement group: {}".format(
                               ray.available_resources(), pg.bundle_specs))

    return pg, bundles


class BaseStrategy:
    """Base class for implementing different placement strategies."""
    placement_group = None
    workers = None

    def create_workers(self):
        raise NotImplementedError

    @property
    def num_workers(self):
        raise NotImplementedError

    @classmethod
    def get_node_workers(cls, workers):
        """Returns list of one worker per node to use for NIC detection."""

        # In some setups (i.e., Peloton), ray nodes may not have
        # unique host names.
        hostnames = map_blocking(lambda w: w.hostname.remote(), workers)
        host_worker_map = {}
        for hostname, worker in zip(hostnames, workers):
            host_worker_map[hostname] = worker

        return list(host_worker_map.values())

    def shutdown(self):
        if self.placement_group:
            ray.util.remove_placement_group(self.placement_group)

        self.workers = []
        self.placement_group = None


class ColocatedStrategy(BaseStrategy):
    """Ensures that the workers are balanced across all hosts."""

    def __init__(self, *, settings, num_hosts: int, num_workers_per_host: int,
                 use_gpu: bool, cpus_per_worker: int, gpus_per_worker: int):
        self.settings = settings
        self.num_hosts = num_hosts
        self.num_workers_per_host = num_workers_per_host
        self.use_gpu = use_gpu
        self.cpus_per_worker = cpus_per_worker
        self.gpus_per_worker = gpus_per_worker or 1

    @property
    def num_workers(self) -> int:
        return int(self.num_hosts * self.num_workers_per_host)

    def _resources_per_host(self):
        num_cpus = self.cpus_per_worker * self.num_workers_per_host
        num_gpus = self.gpus_per_worker * self.num_workers_per_host * int(
            self.use_gpu)
        return dict(CPU=num_cpus, GPU=num_gpus)

    def create_workers(self):
        self.placement_group, bundles = create_placement_group(
            resources_per_bundle=self._resources_per_host(),
            num_bundles=self.num_hosts,
            pg_timeout=self.settings.placement_group_timeout_s,
            pg_strategy="STRICT_SPREAD")

        # Placement group has started. Now create the workers.
        self.workers = []

        # STRICT_SPREAD guarantees each bundle is on a different node.
        # Create num_workers_per_host workers per bundle, i.e. per machine.
        for bundle_index in range(len(bundles)):
            gpu_id_futures = []
            curr_node_workers = []
            remote_cls = ray.remote(BaseHorovodWorker)
            for i in range(self.num_workers_per_host):
                remote_cls_with_options = remote_cls.options(
                    num_cpus=self.cpus_per_worker,
                    num_gpus=self.gpus_per_worker * int(self.use_gpu),
                    placement_group_capture_child_tasks=False,
                    placement_group=self.placement_group,
                    placement_group_bundle_index=bundle_index)
                worker = remote_cls_with_options.remote(
                    world_rank=self.num_workers_per_host * bundle_index + i,
                    world_size=self.num_workers)
                if self.use_gpu:
                    gpu_id_futures.append(worker.get_gpu_ids.remote())
                self.workers.append(worker)
                curr_node_workers.append(worker)
            if len(gpu_id_futures) > 0:
                # By setting CUDA VISIBLE DEVICES to ALL GPUs,
                # CUDA will be able to detect adjacent devices and use IPC
                # allowing for better performance.
                gpu_ids = sum(ray.get(gpu_id_futures), [])
                # Make sure that each worker on the node has unique device.
                assert len(gpu_ids) == len(
                    set(gpu_ids)) == self.num_workers_per_host, gpu_ids
                all_ids = ",".join([str(gpu_id) for gpu_id in gpu_ids])
                futures = []
                for worker in curr_node_workers:
                    futures.append(
                        worker.update_env_vars.remote({
                            "CUDA_VISIBLE_DEVICES":
                            all_ids
                        }))
                ray.get(futures)

        return self.workers, self.get_node_workers(self.workers)


class PGStrategy(BaseStrategy):
    """Packs workers together but does not guarantee balanced hosts.

    Will use the current placement group if one is available.
    """

    def __init__(self, *, settings, num_workers, use_gpu, cpus_per_worker,
                 gpus_per_worker, placement_group=None, force_create_placement_group=False):
        self.settings = settings
        self._num_workers = num_workers
        self.cpus_per_worker = cpus_per_worker
        self.gpus_per_worker = gpus_per_worker or 1
        self.use_gpu = use_gpu
        if force_create_placement_group:
            self.placement_group = None
        else:
            self.placement_group = placement_group or get_current_placement_group()
        self._placement_group_bundles = self.placement_group.bundle_specs if self.placement_group else None
        self._created_placement_group = False

    @property
    def num_workers(self) -> int:
        return int(self._num_workers)

    def resources_per_worker(self):
        num_cpus = self.cpus_per_worker
        num_gpus = self.gpus_per_worker * int(self.use_gpu)
        return dict(CPU=num_cpus, GPU=num_gpus)

    def create_workers(self):
        if not self.placement_group:
            self.placement_group, _ = create_placement_group(
                resources_per_bundle=self.resources_per_worker(),
                num_bundles=self.num_workers,
                pg_strategy="PACK",
                pg_timeout=self.settings.placement_group_timeout_s)
            self._created_placement_group = True

        # Placement group has started. Now create the workers.
        self.workers = []
        remote_cls = ray.remote(BaseHorovodWorker)

        for worker_index in range(self.num_workers):
            remote_cls_with_options = remote_cls.options(
                num_cpus=self.cpus_per_worker,
                num_gpus=self.gpus_per_worker * int(self.use_gpu),
                placement_group_capture_child_tasks=False,
                placement_group=self.placement_group,
                placement_group_bundle_index=worker_index if self._created_placement_group else -1)
            worker = remote_cls_with_options.remote(
                world_rank=worker_index, world_size=self.num_workers)

            self.workers.append(worker)

        if self.use_gpu:
            node_ids = ray.get(
                [worker.node_id.remote() for worker in self.workers])
            gpus = ray.get(
                [worker.get_gpu_ids.remote() for worker in self.workers])
            node_workers = defaultdict(list)
            node_id_to_gpus = defaultdict(list)
            for worker, node_id, worker_gpu_ids in zip(self.workers, node_ids,
                                                       gpus):
                node_workers[node_id].append(worker)
                node_id_to_gpus[node_id].extend(worker_gpu_ids)

            futures = []
            for node_id, gpu_ids in node_id_to_gpus.items():
                all_ids = ",".join([str(gpu_id) for gpu_id in gpu_ids])

                for worker in node_workers[node_id]:
                    futures.append(
                        worker.update_env_vars.remote({
                            "CUDA_VISIBLE_DEVICES":
                            all_ids
                        }))
            ray.get(futures)
        return self.workers, self.get_node_workers(self.workers)

    def shutdown(self):
        if self._created_placement_group and self.placement_group:
            ray.util.remove_placement_group(self.placement_group)
            self.placement_group = None

        self.workers = []



# Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging

from horovod.runner.common.util import codec
from horovod.runner.http.http_server import RendezvousHandler

# GET methods
GET_RANK_AND_SIZE = 'rank_and_size'

# PUT methods
PUT_WORKER_ADDRESSES = 'worker_addresses'


def create_rendezvous_handler(driver):
    class ElasticRendezvousHandler(RendezvousHandler):
        def _get_value(self, scope, key):
            if scope == GET_RANK_AND_SIZE:
                host, local_rank = key.split(':')
                return self._get_rank_and_size(host, int(local_rank))

            return super(RendezvousHandler, self)._get_value(scope, key)

        def _get_rank_and_size(self, host, local_rank):
            logging.info('_get_rank_and_size: {} {}'.format(host, local_rank))
            driver.record_ready(host, local_rank)
            slot_info = driver.get_slot_info(host, local_rank)
            logging.info('rank and size: {} {}'.format(slot_info.rank, slot_info.size))
            return slot_info.to_response_string().encode('ascii')

        def _put_value(self, scope, key, value):
            if scope == PUT_WORKER_ADDRESSES:
                host, local_rank = key.split(':')
                addresses, secret_key = codec.loads_base64(value)
                self._put_worker_addresses(host, int(local_rank), addresses, secret_key)

            super(RendezvousHandler, self)._put_value(scope, key, value)

        def _put_worker_addresses(self, host, local_rank, addresses, secret_key):
            driver.register_worker_server(host, local_rank, addresses, secret_key)

    return ElasticRendezvousHandler



# Copyright (C) 2022, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import contextlib
import horovod.tensorflow.keras as hvd

from horovod.spark.common import constants
from horovod.spark.common.datamodule import DataModule


class PetastormDataModule(DataModule):
    """Default Petastorm-based DataModule for KerasEstimator."""
    def __init__(self, reader_pool_type: str='thread',
                       train_reader_worker_count: int=2,
                       val_reader_worker_count: int=2,
                       make_dataset=None,
                       random_seed=0,
                       **kwargs):
        from petastorm import TransformSpec, make_reader, make_batch_reader

        super().__init__(**kwargs)
        self.reader_pool_type = reader_pool_type
        self.train_reader_worker_count = train_reader_worker_count
        self.val_reader_worker_count = val_reader_worker_count
        self.make_dataset = make_dataset
        self.random_seed = random_seed

        # In general, make_batch_reader is faster than make_reader for reading the dataset.
        # However, we found out that make_reader performs data transformations much faster than
        # make_batch_reader with parallel worker processes. Therefore, the default reader
        # we choose is make_batch_reader unless there are data transformations.
        self.transform_spec = TransformSpec(self.transform_fn) if self.transform_fn else None
        self.reader_factory_kwargs = dict()
        if self.transform_spec:
            self.reader_factory = make_reader
            self.reader_factory_kwargs['pyarrow_serialize'] = True
            self.is_batch_reader = False
        else:
            self.reader_factory = make_batch_reader
            self.is_batch_reader = True

    def __enter__(self):
        super().__enter__()
        self.train_reader = self.reader_factory(self.train_dir,
                                num_epochs=1,
                                cur_shard=hvd.rank(),
                                reader_pool_type=self.reader_pool_type,
                                workers_count=self.train_reader_worker_count,
                                shard_count=hvd.size(),
                                hdfs_driver=constants.PETASTORM_HDFS_DRIVER,
                                schema_fields=self.schema_fields,
                                transform_spec=self.transform_spec,
                                storage_options=self.storage_options,
                                shuffle_rows=self.shuffle,
                                shuffle_row_groups=self.shuffle,
                                seed=self.random_seed,
                                **self.reader_factory_kwargs)
        self.val_reader = self.reader_factory(self.val_dir,
                                num_epochs=1,
                                cur_shard=hvd.rank(),
                                reader_pool_type=self.reader_pool_type,
                                workers_count=self.val_reader_worker_count,
                                shard_count=hvd.size(),
                                hdfs_driver=constants.PETASTORM_HDFS_DRIVER,
                                schema_fields=self.schema_fields,
                                transform_spec=self.transform_spec,
                                storage_options=self.storage_options,
                                shuffle_rows=False,
                                shuffle_row_groups=False,
                                **self.reader_factory_kwargs) if self.has_val else self.empty_batch_reader()
        return self

    def __exit__(self, type, value, traceback):
        if self.has_val and self.val_reader:
            self.val_reader.__exit__(type, value, traceback)
        if self.train_reader:
            self.train_reader.__exit__(type, value, traceback)
        super().__exit__(type, value, traceback)

    @contextlib.contextmanager
    def empty_batch_reader(self):
        yield None

    def train_data(self):
        return self.make_dataset(self.train_reader, self.train_batch_size,
                                 self.is_batch_reader, shuffle=self.shuffle, cache=self.inmemory_cache_all)

    def val_data(self):
        return self.make_dataset(self.val_reader, self.val_batch_size,
                                 self.is_batch_reader, shuffle=False, cache=self.inmemory_cache_all) if self.val_reader else None


class NVTabularDataModule(DataModule):
    """NVTabular-based DataModule for KerasEstimator for GPU-accelerated data loading of tabular datasets.

    Note: requires `label_cols`, `categorical_cols`, and `continuous_cols` to be explicitly provided."""
    def __init__(self, label_cols=[], categorical_cols=[], continuous_cols=[], **kwargs):
        super().__init__(**kwargs)
        self.label_cols = label_cols
        self.categorical_cols = categorical_cols
        self.continuous_cols = continuous_cols
        self.kwargs = kwargs

    @staticmethod
    def seed_fn():
        """
        Generate consistent dataloader shuffle seeds across workers
        Reseeds each worker's dataloader each epoch to get fresh a shuffle
        that's consistent across workers.
        """
        import horovod.tensorflow.keras as hvd
        import numpy as np
        import tensorflow as tf
        hvd.init()
        seed = np.random.randint(0, tf.int32.max)
        seed_tensor = tf.constant(seed)
        root_seed = hvd.broadcast(seed_tensor, name="shuffle_seed", root_rank=0)
        return root_seed

    @staticmethod
    def to_dense(X, labels):
        """Convert NVTabular's ragged representation of array types to dense arrays."""
        import tensorflow as tf
        retX = {}
        for col, tensor in X.items():
            if isinstance(tensor, tuple):
                retX[col] = tf.RaggedTensor.from_row_lengths(values=tensor[0][:,0], row_lengths=tensor[1][:,0]).to_tensor()
            else:
                retX[col] = tensor
        return retX, labels

    def train_data(self):
        import horovod.tensorflow.keras as hvd
        from nvtabular.loader.tensorflow import KerasSequenceLoader, Dataset
        return KerasSequenceLoader(Dataset(self.train_dir, engine="parquet", calculate_divisions=True, **self.kwargs),
                                   batch_size=self.train_batch_size,
                                   label_names=self.label_cols,
                                   cat_names=self.categorical_cols,
                                   cont_names=self.continuous_cols,
                                   engine="parquet",
                                   shuffle=self.shuffle,
                                   buffer_size=0.1,  # how many batches to load at once
                                   parts_per_chunk=1,
                                   global_size=hvd.size(),
                                   global_rank=hvd.rank(),
                                   seed_fn=self.seed_fn).map(self.to_dense)

    def val_data(self):
        import horovod.tensorflow.keras as hvd
        from nvtabular.loader.tensorflow import KerasSequenceLoader, Dataset
        return KerasSequenceLoader(Dataset(self.train_dir, engine="parquet", calculate_divisions=True, **self.kwargs),
                                   batch_size=self.val_batch_size,
                                   label_names=self.label_cols,
                                   cat_names=self.categorical_cols,
                                   cont_names=self.continuous_cols,
                                   engine="parquet",
                                   shuffle=False,
                                   buffer_size=0.06,  # how many batches to load at once
                                   parts_per_chunk=1,
                                   global_size=hvd.size(),
                                   global_rank=hvd.rank()).map(self.to_dense) if self.has_val else self.empty_batch_reader()



# Copyright (C) 2022, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import contextlib
import glob
import horovod.torch as hvd
import os

from horovod.spark.common import constants
from horovod.spark.common.datamodule import DataModule


class PetastormDataModule(DataModule):
    """Default Petastorm-based DataModule for KerasEstimator."""

    def __init__(self,
                 reader_pool_type="thread",
                 train_reader_worker_count=10,
                 val_reader_worker_count=10,
                 random_seed=0,
                 **kwargs):
        from petastorm import TransformSpec, make_reader, make_batch_reader

        super().__init__(**kwargs)
        self.reader_pool_type = reader_pool_type
        self.train_reader_worker_count = train_reader_worker_count
        self.val_reader_worker_count = val_reader_worker_count
        self.random_seed = random_seed

        # In general, make_batch_reader is faster than make_reader for reading the dataset.
        # However, we found out that make_reader performs data transformations much faster than
        # make_batch_reader with parallel worker processes. Therefore, the default reader
        # we choose is make_batch_reader unless there are data transformations.
        self.transform_spec = TransformSpec(self.transform_fn) if self.transform_fn else None
        self.reader_factory_kwargs = dict()
        if self.transform_spec:
            self.reader_factory = make_reader
            self.reader_factory_kwargs['pyarrow_serialize'] = True
        else:
            self.reader_factory = make_batch_reader

    def __enter__(self):
        super().__enter__()
        # Petastorm: read data from the store with the correct shard for this rank
        # setting num_epochs=None will cause an infinite iterator
        # and enables ranks to perform training and validation with
        # unequal number of samples
        self.train_reader = self.reader_factory(
            self.train_dir,
            num_epochs=None,
            cur_shard=hvd.rank(),
            reader_pool_type=self.reader_pool_type,
            workers_count=self.train_reader_worker_count,
            shard_count=hvd.size(),
            hdfs_driver=constants.PETASTORM_HDFS_DRIVER,
            schema_fields=self.schema_fields,
            transform_spec=self.transform_spec,
            storage_options=self.storage_options,
            shuffle_rows=self.shuffle,
            shuffle_row_groups=self.shuffle,
            seed=self.random_seed,
            **self.reader_factory_kwargs
        )

        self.val_reader = self.reader_factory(
            self.val_dir,
            num_epochs=None,
            cur_shard=hvd.rank(),
            reader_pool_type=self.reader_pool_type,
            workers_count=self.val_reader_worker_count,
            shard_count=hvd.size(),
            hdfs_driver=constants.PETASTORM_HDFS_DRIVER,
            schema_fields=self.schema_fields,
            transform_spec=self.transform_spec,
            storage_options=self.storage_options,
            shuffle_rows=False,
            shuffle_row_groups=False,
            seed=self.random_seed,
            **self.reader_factory_kwargs
        ) if self.has_val else self.empty_batch_reader()

        return self

    def __exit__(self, type, value, traceback):
        if self.has_val and self.val_reader:
            self.val_reader.__exit__(type, value, traceback)
        if self.train_reader:
            self.train_reader.__exit__(type, value, traceback)
        super().__exit__(type, value, traceback)

    @contextlib.contextmanager
    def empty_batch_reader(self):
        yield None

    def train_data(self):
        from petastorm.pytorch import BatchedDataLoader, InMemBatchedDataLoader

        if self.inmemory_cache_all:
            train_loader = InMemBatchedDataLoader(self.train_reader,
                                                  batch_size=self.train_batch_size,
                                                  num_epochs=self.num_train_epochs,
                                                  rows_capacity=self.steps_per_epoch_train*self.train_batch_size,
                                                  shuffle=self.shuffle)
        else:
            train_loader = BatchedDataLoader(self.train_reader,
                                             batch_size=self.train_batch_size,
                                             # No need to shuffle again in dataloader level
                                             shuffling_queue_capacity=0)
        return train_loader

    def val_data(self):
        from petastorm.pytorch import BatchedDataLoader, InMemBatchedDataLoader

        if self.inmemory_cache_all:
            val_loader = InMemBatchedDataLoader(self.val_reader,
                                                batch_size=self.val_batch_size,
                                                num_epochs=self.num_train_epochs,
                                                rows_capacity=self.steps_per_epoch_val*self.val_batch_size,
                                                shuffle=False)
        else:
            val_loader = BatchedDataLoader(self.val_reader,
                                           batch_size=self.val_batch_size,
                                           shuffling_queue_capacity=0)
        return val_loader


class MapIterable():
    """Wraps an iterable with a user-defined map function for N epochs."""

    def __init__(self, data, epochs=None, map_fn=lambda x: x):
        self.data = data
        self.epochs = epochs
        self.map_fn = map_fn

    def __iter__(self):
        if self.epochs:
            for _ in range(self.epochs):
                for x in self.data:
                    yield self.map_fn(x)
        else:
            while True:
                for x in self.data:
                    yield self.map_fn(x)


class NVTabularDataModule(DataModule):
    """NVTabular-based DataModule for TorchEstimator for GPU-accelerated data loading of tabular datasets.

    Note: requires `label_cols`, `categorical_cols`, and `continuous_cols` to be explicitly provided."""

    def __init__(self, label_cols=[], categorical_cols=[], continuous_cols=[], **kwargs):
        super().__init__(**kwargs)
        self.label_cols = label_cols
        self.categorical_cols = categorical_cols
        self.continuous_cols = continuous_cols
        self.kwargs = kwargs

    @staticmethod
    def seed_fn():
        """
        Generate consistent dataloader shuffle seeds across workers
        Reseeds each worker's dataloader each epoch to get fresh a shuffle
        that's consistent across workers.
        """
        import numpy as np
        import torch
        hvd.init()
        seed = np.random.randint(0, torch.iinfo(torch.int32).max)
        seed_tensor = torch.tensor(seed)
        root_seed = hvd.broadcast(seed_tensor, name="shuffle_seed", root_rank=0)
        return root_seed

    def _transform(self, features_and_label):
        """Transform NVTabular value-and-offsets arrays into torch arrays."""
        import torch
        features, label = features_and_label
        for k, v in features.items():
            if isinstance(v, tuple):  # values and offsets
                indices = v[1].flatten().tolist()
                features[k] = torch.vstack(v[0].tensor_split(indices[1:]))
        return features, label

    def train_data(self):
        import nvtabular as nvt
        from nvtabular.loader.torch import TorchAsyncItr, DLDataLoader

        train_dataset = TorchAsyncItr(
            nvt.Dataset(self.train_dir, engine='parquet', calculate_divisions=True, **self.kwargs),
            batch_size=self.train_batch_size,
            cats=self.categorical_cols,
            conts=self.continuous_cols,
            labels=self.label_cols,
            shuffle=self.shuffle,
            parts_per_chunk=1,
            global_size=hvd.size(),
            global_rank=hvd.rank(),
            seed_fn=self.seed_fn)

        train_dataloader = DLDataLoader(train_dataset, batch_size=None, collate_fn=lambda x: x, pin_memory=False, num_workers=0)
        return MapIterable(train_dataloader, epochs=self.num_train_epochs, map_fn=self._transform)

    def val_data(self):
        import nvtabular as nvt
        from nvtabular.loader.torch import TorchAsyncItr, DLDataLoader

        val_dataset = TorchAsyncItr(
            nvt.Dataset(self.val_dir, engine='parquet', calculate_divisions=True, **self.kwargs),
            batch_size=self.val_batch_size,
            cats=self.categorical_cols,
            conts=self.continuous_cols,
            labels=self.label_cols,
            shuffle=False,
            parts_per_chunk=1,
            global_size=hvd.size(),
            global_rank=hvd.rank()) if self.has_val else None

        val_dataloader = DLDataLoader(val_dataset, batch_size=None, collate_fn=lambda x: x, pin_memory=False, num_workers=0)
        return MapIterable(val_dataloader, epochs=self.num_train_epochs, map_fn=self._transform)
